// Forlix SrcDS Manager
// http://forlix.org/, df@forlix.org
//
// Copyright (c) 2010-2013 Dominik Friedrichs

#include "SrcDS_Manager.h"


static void WaitForNextTick(time_t *tnext)
{
  while(time(NULL) < *tnext)
    Sleep(20);

  *tnext = time(NULL)+TICK_INTERVAL;
  return;
}


static bool MonitorShutdown()
{
  MSG msg;

  while(PeekMessage(&msg, NULL, WM_APP, WM_APP+0xFF, PM_REMOVE))
  switch(msg.message)
  // always check for messages
  {
    case WM_THREAD_TERMINATE:
      return(true);
    break;
  }

  return(false);
}


DWORD WINAPI MonitorThread(LPVOID lpParam)
{
  const MT_ServerData *mtsd = (MT_ServerData*)lpParam;

  PROCESS_INFORMATION pi_srv;
  PROCESS_INFORMATION pi_upd;
  memset(&pi_srv, 0, sizeof(pi_srv));
  memset(&pi_upd, 0, sizeof(pi_upd));

  // build update marker path
  char umarker[MAX_PATH];
  sprintf(umarker, "%s\\%s\\%s", mtsd->Path, mtsd->GameDir, MRR_MARKERFILE);

  enum MonitorState {LaunchServer, ServerRunning,
                     LaunchUpdate,
                     LaunchUpdate_Tool, ToolUpdateRunning,
                     LaunchUpdate_Main, MainUpdateRunning};

  MonitorState mstate = LaunchServer;
  time_t mdelay = 0;

  unsigned int pingfail = 0;

  HANDLE update_pipe = NULL;
  CStrBuf update_log(256);
  unsigned int update_try;
  time_t update_timeout;

  if(FileExists(umarker)
  || !FolderExists(mtsd->Path))
  // update pending or game path missing
    mstate = LaunchUpdate;

  for(time_t ticknext = 0; !MonitorShutdown(); WaitForNextTick(&ticknext))
  if(time(NULL) >= mdelay)
  switch(mstate)
  // tick loop
  {
    case LaunchServer:
    {
      if(!CreateServerProcess(mtsd, &pi_srv))
      {
        LogMsg("CreateServerProcess failed for \"%s\", retry in 10s", mtsd->FriendlyName);

        // try again in 10 seconds
        mdelay = time(NULL)+10;
        break;
      }

      LogMsg("\"%s\" launched successfully", mtsd->FriendlyName);

      mstate = ServerRunning;
      mdelay = time(NULL)+strtoul(mtsd->MonitorDelay, NULL, 10);
    }
    break;

    case ServerRunning:
    {
      ServerHealth health = GetServerHealth(mtsd, &pi_srv, &pingfail);

      if(health == Server_Healthy)
      // re-check in 5 seconds
      {
        mdelay = time(NULL)+5;
        break;
      }

      // make sure its really gone
      TerminateProcess(pi_srv.hProcess, 1);

      switch(health)
      {
        case Server_MemoryFull:
          LogMsg("\"%s\" exceeded memory limit and has been terminated", mtsd->FriendlyName);
        break;

        case Server_TimedOut:
          LogMsg("\"%s\" was unresponsive and has been terminated", mtsd->FriendlyName);
        break;

        case Server_Terminated:
          LogMsg("\"%s\" has terminated", mtsd->FriendlyName);
        break;
      }

      CloseHandle(pi_srv.hThread);
      CloseHandle(pi_srv.hProcess);

      if(FileExists(umarker))
      {
        LogMsg("\"%s\" requested an update", mtsd->FriendlyName);
        mstate = LaunchUpdate;
      }
      else
        mstate = LaunchServer;

      pingfail = 0;
      mdelay = time(NULL)+5;
    }
    break;

    case LaunchUpdate:
    {
      if(WaitForSingleObject(g_UpdMutex, 250) != WAIT_OBJECT_0)
      // allow only one server to update at a time
      {
        mdelay = time(NULL)+10;
        break;
      }

      update_try = 0;
      mstate = LaunchUpdate_Tool;
    }
    break;

    case LaunchUpdate_Tool:
    {
      if(!CreateUpdateProcess(mtsd, &pi_upd, update_pipe, true))
      {
        LogMsg("[Update] CreateUpdateProcess failed for \"%s\", retry in 10s", mtsd->FriendlyName);

        // try again in 10 seconds
        mdelay = time(NULL)+10;
        break;
      }

      LogMsg("[Update] Tool version check");
      update_timeout = time(NULL)+TOOLUPDATE_TIMEOUT;

      mstate = ToolUpdateRunning;
    }
    break;

    case ToolUpdateRunning:
    {
      // read pipe contents into growing buffer
      FetchFromPipe(update_pipe, &update_log);

      if(!HasProcessTerminated(pi_upd.hProcess))
      // wait for update process to terminate
      {
        if(time(NULL) >= update_timeout)
        {
          TerminateProcess(pi_upd.hProcess, 1);
          CloseHandle(pi_upd.hThread);
          CloseHandle(pi_upd.hProcess);

          LogMsg("[Update] Timed out, retry in 120s");

          CloseHandle(update_pipe);
          update_log.Wipe();

          mstate = LaunchUpdate_Tool;
          mdelay = time(NULL)+120;
        }

        break;
      }

      const char *rptr = update_log.Read();
      char splitbuf[64];
      bool toolupdate = false;

      while(strsplt(splitbuf, sizeof(splitbuf), "\r\n", rptr) != NULL)
      if(strstr(splitbuf, "Downloading update"))
        toolupdate = true;

      CloseHandle(update_pipe);
      update_log.Wipe();

      if(toolupdate)
      {
        LogMsg("[Update] Tool has been updated");
        mstate = LaunchUpdate_Tool;
      }
      else
      {
        LogMsg("[Update] Tool is up to date");
        mstate = LaunchUpdate_Main;
      }

      mdelay = time(NULL)+5;
    }
    break;

    case LaunchUpdate_Main:
    {
      // now launch main update
      if(!CreateUpdateProcess(mtsd, &pi_upd, update_pipe, false))
      {
        LogMsg("[Update] CreateUpdateProcess failed for \"%s\", retry in 10s", mtsd->FriendlyName);

        // try again in 10 seconds
        mdelay = time(NULL)+10;
        break;
      }

      LogMsg("[Update] Launched for \"%s\"", mtsd->FriendlyName);
      update_timeout = time(NULL)+MAINUPDATE_TIMEOUT;

      update_try++;
      mstate = MainUpdateRunning;
    }
    break;

    case MainUpdateRunning:
    {
      // read pipe contents into growing buffer
      FetchFromPipe(update_pipe, &update_log);

      if(!HasProcessTerminated(pi_upd.hProcess))
      // wait for update process to terminate and the pipe to be empty
      {
        if(time(NULL) >= update_timeout)
        {
          TerminateProcess(pi_upd.hProcess, 1);
          CloseHandle(pi_upd.hThread);
          CloseHandle(pi_upd.hProcess);

          LogMsg("[Update] Timed out, retry in 120s");

          WriteUpdateLog(mtsd, update_log.Read());

          CloseHandle(update_pipe);
          update_log.Wipe();

          mstate = LaunchUpdate_Main;
          mdelay = time(NULL)+120;
        }

        break;
      }

      WriteUpdateLog(mtsd, update_log.Read());

      const char *rptr = update_log.Read();
      char splitbuf[64];
      bool success = false;

      // look for the success-line
      while(strsplt(splitbuf, sizeof(splitbuf), "\r\n", rptr) != NULL)
      if(strstr(splitbuf, "Success!"))
        success = true;

      CloseHandle(update_pipe);
      update_log.Wipe();

      if(!success)
      // result not reliable
      {
        if(update_try >= 3)
        {
          LogMsg("[Update] Ran %u times without reliable result", update_try);

          DeleteFile(umarker);

          ReleaseMutex(g_UpdMutex);

          mstate = LaunchServer;
          mdelay = time(NULL)+10;
          break;
        }

        LogMsg("[Update] Result unreliable, retry in 10s");

        mstate = LaunchUpdate_Main;
        mdelay = time(NULL)+10;
        break;
      }

      LogMsg("[Update] Completed successfully");

      // update successful - delete marker and finally release mutex
      DeleteFile(umarker);

      ReleaseMutex(g_UpdMutex);

      mstate = LaunchServer;
      mdelay = time(NULL)+5;
    }
    break;
  }

  // make sure we leave nothing behind
  if(TerminateProcess(pi_srv.hProcess, 0))
    LogMsg("\"%s\" has been terminated", mtsd->FriendlyName);

  CloseHandle(update_pipe);

  CloseHandle(pi_srv.hThread);
  CloseHandle(pi_srv.hProcess);

  if(TerminateProcess(pi_upd.hProcess, 0))
    LogMsg("[Update] Terminated for \"%s\"", mtsd->FriendlyName);

  CloseHandle(pi_upd.hThread);
  CloseHandle(pi_upd.hProcess);

  return(0);
}
